"""Reachy main module entry point."""

import time
import logging
import numpy as np

from operator import attrgetter

from .parts import LeftArm, RightArm, Head


logger = logging.getLogger(__name__)


class Reachy(object):
    """Class representing the connection with the hardware robot.

    Args:
        left_arm (reachy.parts.LeftArm): left arm part if present or None if absent
        right_arm (reachy.parts.RightArm): right arm part if present or None if absent
        head (reachy.parts.Head): hrad part if present or None if absent

    Connect and synchronize with the hardware robot.

    It can be used to monitor real time robot state and to send commands.
    Mainly a container to hold the different parts of Reachy together.
    """

    def __init__(self,
                 left_arm=None,
                 right_arm=None,
                 head=None):
        """
        Connect and synchronize with the hardware robot.

        Args:
            left_arm (reachy.parts.LeftArm): left arm part if present or None if absent
            right_arm (reachy.parts.RightArm): right arm part if present or None if absent
            head (reachy.parts.Head): hrad part if present or None if absent
        """
        self._parts = []

        if left_arm is not None:
            if not isinstance(left_arm, LeftArm):
                raise ValueError('"left_arm" must be a LeftArm or None!')
            self._parts.append(left_arm)
        self.left_arm = left_arm

        if right_arm is not None:
            if not isinstance(right_arm, RightArm):
                raise ValueError('"right_arm" must be a RightArm or None!')
            self._parts.append(right_arm)
        self.right_arm = right_arm

        if head is not None:
            if not isinstance(head, Head):
                raise ValueError('"head" must be a Head or None!')
            self._parts.append(head)
        self.head = head

        logger.info(
            'Connected to reachy',
            extra={
                'parts': [p.name for p in self.parts],
            }
        )

    def __repr__(self):
        """Reachy representation."""
        return f'<Reachy: {self.parts}>'

    def close(self):
        """Close all communication with each attached parts."""
        logger.info(
            'Closing connection with reachy',
            extra={
                'parts': [p.name for p in self.parts],
            }
        )
        for p in self.parts:
            p.teardown()

    @property
    def parts(self):
        """List of all attached parts."""
        return self._parts

    @property
    def motors(self):
        """List of all motors in the attached parts."""
        return sum([p.motors for p in self.parts], [])

    def goto(self,
             goal_positions, duration,
             starting_point='present_position',
             wait=False, interpolation_mode='linear'):
        """
        Goto specified goal positions.

        Args:
            goal_positions (dict): desired target position (in the form {'full_motor_name': target_position})
            duration (float): move duration (in sec.)
            starting_point (str): register to use to retrieve the starting point (e.g. 'present_postion' or 'goal_position')
            wait (bool): whether or not to wait for the end motion before returning
            interpolation_mode (str): interpolation used for computing the trajectory (e.g. 'linear' or 'minjerk')

        Returns:
            list: list of reachy.trajectory.TrajectoryPlayer (one for each controlled motor)

        """
        trajs = []

        for i, (motor_name, goal_pos) in enumerate(goal_positions.items()):
            last = wait and (i == len(goal_positions) - 1)

            motor = attrgetter(motor_name)(self)
            trajs.append(motor.goto(goal_pos, duration, starting_point,
                         wait=last, interpolation_mode=interpolation_mode))

        return trajs

    def need_cooldown(self, temperature_limit=50):
        """
        Check if Reachy needs to cool down.

        Args:
            temperature_limit (int): Temperature limit (in °C) for each motor.

        Returns:
            bool: Whether or not you should let the robot cool down

        """
        motor_temperature = np.array([
            m.temperature for m in self.motors
        ])
        logger.info(
            'Checking Reachy motors temperature',
            extra={
                'temperatures': {
                    m.name: m.temperature for m in self.motors
                }
            }
        )
        return np.any(motor_temperature > temperature_limit)

    def wait_for_cooldown(self, rest_position, goto_rest_duration=5, lower_temperature=45):
        """
        Wait for the robot to lower its temperature.

        The robot will first go to the specified rest position and then, it will turn all motors compliant.
        Finally, it will wait until the temperature of each motor goes below the lower_temperature parameters.

        Args:
            rest_position (dict): the desired rest position for the robot
            goto_rest_duration (float): time in seconds to reach the rest position
            lower_temeprature (int): lower temperature threshold (in °C) to be reached by all motors before the end of cool down

        .. note:: The robot will stay compliant at the end of the function call.
                  It is up to you, to put it back in the desired position.

        """
        self.goto(
            goal_positions=rest_position,
            duration=goto_rest_duration,
            wait=True,
        )

        for m in self.motors:
            m.compliant = True

        while True:
            motor_temperature = np.array([
                m.temperature for m in self.motors
            ])

            logger.warning(
                'Motors cooling down...',
                extra={
                    'temperatures': {
                        m.name: m.temperature for m in self.motors
                    }
                },
            )

            if np.all(motor_temperature < lower_temperature):
                break

            time.sleep(30)



"""Command line utility tool to setup all dynamixel motors angle limits in a Reachy robot.

.. warning:: Make sure the motor ids are correctly set!
"""

import time
import argparse

from reachy.parts import RightArm, LeftArm, Head
from reachy.parts.hand import LeftForceGripper, RightForceGripper
from reachy.error import LuosModuleNotFoundError, LuosGateNotFoundError


reachy_config = {}
reachy_config.update({f'right_arm.{m}': c for m, c in RightArm.dxl_motors.items()})
reachy_config.update({f'right_arm.hand.{m}': c for m, c in RightForceGripper.dxl_motors.items()})
reachy_config.update({f'left_arm.{m}': c for m, c in LeftArm.dxl_motors.items()})
reachy_config.update({f'left_arm.hand.{m}': c for m, c in LeftForceGripper.dxl_motors.items()})
reachy_config.update({f'head.{m}': c for m, c in Head.dxl_motors.items()})


def main():
    """Configure all dynamixel motor limits on a Reachy robot."""
    parser = argparse.ArgumentParser()
    parser.add_argument('--luos_port', default='/dev/ttyUSB*')
    args = parser.parse_args()

    for Arm in (LeftArm, RightArm):
        part_name = Arm.__name__

        print(f'Connecting to part "{part_name}"...')
        try:
            arm = Arm(io=args.luos_port, hand='force_gripper')
            time.sleep(0.25)

            for m in arm.motors:
                limits = reachy_config[m.name]['angle-limits']
                print(f'\t Setting motor "{m.name}" limits to {limits}.')
                m._motor.rot_position_limit = tuple(limits)
                time.sleep(0.25)

        except LuosModuleNotFoundError as e:
            print(f'Could not connect to {part_name}')
            print(f'Error: {e}')
        except LuosGateNotFoundError:
            print(f'No {part_name} found, skipping.')


if __name__ == '__main__':
    main()



"""Wrapper around two RPI and the multi channel switch.

It relies on the use of the Multi_Adapter_Board_2Channel_uc444 for switching camera via the I2C and GPIO control.

See https://github.com/ArduCAM/RaspberryPi/tree/master/Multi_Camera_Adapter/Multi_Adapter_Board_2Channel_uc444

"""

import time
import cv2 as cv

from threading import Thread, Event, Lock

from ..error import CameraNotFoundError


class BackgroundVideoCapture(object):
    """Wrapper on OpenCV VideoCapture object.

    Args:
        camera_index (int): index of the used camera (see OpenCV doc for details)
        resolution (int, int): desired resolution for the grabbed frame (the resolution must be compatible with the driver)

    Instantiating this object will automatically start the polling of image in background.

    This wrapper is reponsible for automatically polling image on the camera.
    This ensures that we can always access the most recent image.
    """

    def __init__(self, camera_index, resolution=(600, 800), lazy_setup=True):
        """Open video capture on the specified camera."""
        self.camera_index = camera_index
        self.resolution = resolution

        if not lazy_setup:
            self._setup()

    def _setup(self):
        self.cap = cv.VideoCapture(self.camera_index)

        if not self.cap.isOpened():
            raise CameraNotFoundError(
                message=f'Camera {self.camera_index} not found!',
                camera_id=self.camera_index,
            )

        self.cap.set(cv.CAP_PROP_FOURCC, cv.VideoWriter_fourcc('M', 'J', 'P', 'G'))
        self.cap.set(cv.CAP_PROP_FRAME_HEIGHT, self.resolution[0])
        self.cap.set(cv.CAP_PROP_FRAME_WIDTH, self.resolution[1])

        self._lock = Lock()
        self.running = Event()

        self._img = None

        self._t = Thread(target=self._read_loop)
        self._t.daemon = True
        self._t.start()

        for _ in range(50):
            time.sleep(0.1)
            if self._img is not None:
                break

    def close(self):
        """Stop polling image and release the Video Capture."""
        if hasattr(self, 'cap'):
            self.running.clear()

            if self._t.is_alive():
                self._t.join()

            self.cap.release()

    def _read_loop(self):
        self.running.set()

        while self.running.is_set():
            b, img = self.cap.read()

            if b:
                with self._lock:
                    self._img = img.copy()

    def read(self):
        """Retrieve the last grabbed image."""
        if not hasattr(self, 'cap'):
            self._setup()

        with self._lock:
            return self._img is not None, self._img



"""Abstract IO definition."""


class IO(object):
    """Abstract IO class.

    Defines how to find and interact with the module.
    """

    def find_module(self, module_name):
        """Get a specific module from the IO."""
        raise NotImplementedError

    def find_dxl(self, dxl_name, dxl_config):
        """Get a specific dynamixel motor from the IO."""
        raise NotImplementedError

    def find_fan(self, fan_name):
        """Get a specific fan from its name."""
        raise NotImplementedError

    def find_orbita_disks(self):
        """Get a specific orbita disk from the IO."""
        raise NotImplementedError

    def close(self):
        """Close and clean the IO."""
        raise NotImplementedError

    def find_camera(self, camera_index):
        """Return a camera.

        The camera object must implement following methods:
            * read() -> bool, img
            * close()
        """
        raise NotImplementedError



"""Arm part module.

Implements a Right and a Left Arm.
"""

import time
import numpy as np

from operator import attrgetter
from collections import OrderedDict
from threading import Thread, Event

from .hand import LeftEmptyHand, RightEmptyHand, LeftForceGripper, RightForceGripper, OrbitaWrist
from .part import ReachyPart


hands = {
    'force_gripper': {'left': LeftForceGripper, 'right': RightForceGripper},
    'orbita_wrist': {'left': OrbitaWrist, 'right': OrbitaWrist},
    'empty_hand': {'left': LeftEmptyHand, 'right': RightEmptyHand},
}


class Arm(ReachyPart):
    """Arm abstraction class.

    Args:
        side (str): 'right' or 'left'
        io (str): port name where the modules can be found
        dxl_motors (dict): config of the dynamixel motors composing the arm
        hand (str): name of the Hand to attach ('force_gripper', 'orbita_wrist' or it can be None if no hand are attached)

    Provides high-level access to:
        * ordered list of motors
        * forward and inverse kinematics
    """

    fans = {
        'shoulder_fan': 'shoulder_pitch',
        'elbow_fan': 'elbow_pitch',
    }
    lower_temp_threshold, upper_temp_threshold = 40, 45

    def __init__(self, side, io, dxl_motors, hand):
        """Create a new Arm part."""
        ReachyPart.__init__(self, name=f'{side}_arm', io=io)
        self.side = side

        dxl_motors = OrderedDict(dxl_motors)

        self.attach_dxl_motors(dxl_motors)

        if hand is not None and hand not in hands.keys():
            raise ValueError(f'"hand" must be one of {list(hands.keys())} or None!')

        if hand is not None:
            hand_cls = hands[hand][side]

            hand_part = hand_cls(root=self, io=io)
            self.motors += hand_part.motors
            self.hand = hand_part

            for m, conf in hand_cls.dxl_motors.items():
                dxl_motors[m] = conf

        else:
            self.hand = None

        self.attach_kinematic_chain(dxl_motors)

        self.fans = dict(Arm.fans)
        if hand is not None:
            self.fans.update(hand_cls.fans)

        for name in self.fans.keys():
            setattr(self, name, self.io.find_fan(name))

        self._monitor_temp = Event()
        self._monitor_temp_loop = None
        self.enable_temperature_monitoring()

    def teardown(self):
        """Clean up before closing."""
        if self.hand is not None:
            self.hand.teardown()
        ReachyPart.teardown(self)

    def __repr__(self):
        """Arm representation."""
        return f'<{self.side.capitalize()}Arm "motors": {self.motors} "hand": {self.hand}>'

    def forward_kinematics(self, joints_position, use_rad=False):
        """Compute the forward kinematics of the Arm.

        Args:
            joints_position (:py:class:`~numpy.ndarray`): angle joints configuration of the arm (in degrees by default)
            use_rad (bool): whether or not to use radians for joints configuration

        .. note:: the end effector will be the end of the Hand if one is attached.

        """
        joints_position = np.array(joints_position)
        if len(joints_position.shape) == 1:
            joints_position = joints_position.reshape(1, -1)

        if not use_rad:
            joints_position = np.deg2rad(joints_position)

        M = self.kin_chain.forward(joints_position)

        if joints_position.shape[0] == 1:
            M = M[0]
        return M

    def inverse_kinematics(self, target_pose, q0=None, use_rad=False, maxiter=10):
        """Approximate the inverse kinematics of the Arm.

        Args:
            target_pose (:py:class:`~numpy.ndarray`): 4x4 homogeneous pose of the target end effector pose
            q0 (:py:class:`~numpy.ndarray`): joint initial angle configurations (used for bootstraping the optimization)
            use_rad (bool): whether or not to use radians for joints configuration
            maxiter (int): maximum number of iteration to run on the optimizer

        .. note:: the end effector will be the end of the Hand if one is attached.

        """
        if q0 is None:
            q0 = [m.present_position for m in self.motors]

        q0 = np.array(q0)

        if len(q0.shape) == 1:
            q0 = q0.reshape(1, -1)

        if len(target_pose.shape) == 2:
            target_pose = target_pose.reshape(-1, 4, 4)

        if not use_rad:
            q0 = np.deg2rad(q0)

        J = self.kin_chain.inverse(target_pose, q0, maxiter=maxiter)

        if J.shape[0] == 1:
            J = J[0]

        if not use_rad:
            J = np.rad2deg(J)

        return J

    def enable_temperature_monitoring(self):
        """Enable the automatic motor cooling procedure.

        The specified motors temperature will be watched and when they reached a specific threshold, the fan will automatically be turned on.
        When the temperature goes below a lower threshold, they will turn off.
        """
        if not self._monitor_temp.is_set():
            self._monitor_temp.set()
            self._monitor_temp_loop = Thread(target=self._temperature_monitoring)
            self._monitor_temp_loop.daemon = True
            self._monitor_temp_loop.start()

    def disable_temperature_monitoring(self):
        """Disable the automatic motor cooling procedure."""
        if self._monitor_temp.is_set():
            self._monitor_temp.clear()

    def _temperature_monitoring(self):
        while self._monitor_temp.is_set():
            for fan_name, motor_name in self.fans.items():
                fan = attrgetter(fan_name)(self)
                motor = attrgetter(motor_name)(self)

                if motor.temperature is not None and motor.temperature >= self.upper_temp_threshold:
                    fan.on()
                elif motor.temperature is not None and motor.temperature <= self.lower_temp_threshold:
                    fan.off()

            time.sleep(30)


class LeftArm(Arm):
    """Left Arm part.

    Args:
        io (str): port name where the modules can be found
        hand (str): name of the :py:class:`~reachy.parts.hand.Hand` to attach ('force_gripper', 'orbita_wrist' or it can be None if no hand are attached)
    """

    dxl_motors = OrderedDict([
        ('shoulder_pitch', {
            'id': 20, 'offset': 90.0, 'orientation': 'direct',
            'angle-limits': [-60, 180],
            'link-translation': [0, 0.19, 0], 'link-rotation': [0, 1, 0]
        }),
        ('shoulder_roll', {
            'id': 21, 'offset': -90.0, 'orientation': 'indirect',
            'angle-limits': [-90, 100],
            'link-translation': [0, 0, 0], 'link-rotation': [1, 0, 0],
        }),
        ('arm_yaw', {
            'id': 22, 'offset': 0.0, 'orientation': 'indirect',
            'angle-limits': [-90, 90],
            'link-translation': [0, 0, 0], 'link-rotation': [0, 0, 1],
        }),
        ('elbow_pitch', {
            'id': 23, 'offset': 0.0, 'orientation': 'indirect',
            'angle-limits': [0, 125],
            'link-translation': [0, 0, -0.28], 'link-rotation': [0, 1, 0],
        }),
    ])

    def __init__(self, io, hand=None):
        """Create a new Left Arm part."""
        Arm.__init__(self, side='left',
                     io=io, dxl_motors=LeftArm.dxl_motors,
                     hand=hand)


class RightArm(Arm):
    """Right Arm part.

    Args:
        io (str): port name where the modules can be found
        hand (str): name of the :py:class:`~reachy.parts.hand.Hand` to attach ('force_gripper', 'orbita_wrist' or it can be None if no hand are attached)
    """

    dxl_motors = OrderedDict([
        ('shoulder_pitch', {
            'id': 10, 'offset': 90.0, 'orientation': 'indirect',
            'angle-limits': [-180, 60],
            'link-translation': [0, -0.19, 0], 'link-rotation': [0, 1, 0],
        }),
        ('shoulder_roll', {
            'id': 11, 'offset': 90.0, 'orientation': 'indirect',
            'angle-limits': [-100, 90],
            'link-translation': [0, 0, 0], 'link-rotation': [1, 0, 0],
        }),
        ('arm_yaw', {
            'id': 12, 'offset': 0.0, 'orientation': 'indirect',
            'angle-limits': [-90, 90],
            'link-translation': [0, 0, 0], 'link-rotation': [0, 0, 1],
        }),
        ('elbow_pitch', {
            'id': 13, 'offset': 0.0, 'orientation': 'indirect',
            'angle-limits': [0, 125],
            'link-translation': [0, 0, -0.28], 'link-rotation': [0, 1, 0],
        }),
    ])

    def __init__(self, io, hand=None):
        """Create a new Right Arm part."""
        Arm.__init__(self, side='right',
                     io=io, dxl_motors=RightArm.dxl_motors,
                     hand=hand)



"""Trajectory interpolation utility module.

This module defines various interpolation technique (linear, minimum jerk).
They can be used in all goto functions.
"""
import time
import numpy as np

from threading import Thread, Event
from scipy.interpolate import interp1d


class TrajectoryInterpolation(object):
    """Trajectory interpolation abstraction class.

    Args:
        initial_position (float): starting position (in degrees)
        goal_position (float): end position (in degrees)
        duration (float): duration of the movement (in seconds)

    You can defined your own interpolation technique by respecting this abstraction so they can be used in goto functions.
    """

    def __init__(self, initial_position, goal_position, duration):
        """Create your interpolation object."""
        self.initial_position = initial_position
        self.goal_position = goal_position
        self.duration = duration

        self._t = None
        self._running = Event()

    def interpolate(self, t):
        """Interpolate the position at given time.

        Args:
            t (float): time where to interpolate

        You are responsible for implementing this method in your own interpolation technique.
        Please refer to the implementation of Linear of MinimumJerk for examples.
        """
        raise NotImplementedError

    def start(self, motor, update_freq=100):
        """Start the interpolation trajectory thread.

        Args:
            motor (motor): motor to apply the trajectory to
            update_freq (float): Update sample frequency (in Hz)
        """
        self._t = Thread(target=lambda: self._follow_traj_loop(motor, update_freq))
        self._running.set()
        self._t.start()

    @property
    def is_playing(self):
        """Check if the trajectory is currently playing."""
        return self._t is not None and self._t.is_alive()

    def stop(self, wait=True):
        """Stop the interpolation trajectory."""
        self._running.clear()
        if wait:
            self.wait()

    def wait(self):
        """Block until the end of the trajectory interpolation."""
        if self.is_playing:
            self._t.join()

    def _follow_traj_loop(self, motor, update_freq):
        t0 = time.time()

        while self._running.is_set():
            t = time.time() - t0
            if t > self.duration:
                break

            pos = self.interpolate(t)
            if hasattr(motor, 'goal_position'):
                motor.goal_position = pos
            else:
                motor.target_rot_position = pos

            time.sleep(1 / update_freq)


class Linear(TrajectoryInterpolation):
    """Linear implementation implementation."""

    def interpolate(self, t):
        """Linear interpolation at time t."""
        return self.initial_position + (self.goal_position - self.initial_position) * t / self.duration


class MinimumJerk(TrajectoryInterpolation):
    """Minimum Jerk interpolation implementation.

    Args:
        initial_position (float): starting position (in degrees)
        goal_position (float): end position (in degrees)
        duration (float): duration of the movement (in seconds)
        initial_velocity (float): initial velocity used for interpolation
        final_velocity (float): final velocity used for interpolation
        initial_acceleration (float): initial acceleration used for interpolation
        final_acceleration (float): final acceleration used for interpolation
    """

    def __init__(
        self,
        initial_position, goal_position, duration,
        initial_velocity=0, final_velocity=0, initial_acceleration=0, final_acceleration=0,
    ):
        """Create the minjerk interpolation."""
        TrajectoryInterpolation.__init__(self, initial_position, goal_position, duration)

        a0 = initial_position
        a1 = initial_velocity
        a2 = initial_acceleration / 2

        d1, d2, d3, d4, d5 = [duration ** i for i in range(1, 6)]

        A = np.array((
            (d3, d4, d5),
            (3 * d2, 4 * d3, 5 * d4),
            (6 * d1, 12 * d2, 20 * d3)
        ))
        B = np.array((
            goal_position - a0 - (a1 * d1) - (a2 * d2),
            final_velocity - a1 - (2 * a2 * d1),
            final_acceleration - (2 * a2)
        ))
        X = np.linalg.solve(A, B)

        self._coeffs = [
            a0,
            a1,
            a2,
            X[0],
            X[1],
            X[2]
        ]

    def interpolate(self, t):
        """Minjerk interpolation at time t."""
        return np.sum([
            c * t ** i
            for i, c in enumerate(self._coeffs)
        ], axis=0)


def cubic_smooth(traj, nb_kp, out_points=-1):
    """Trjaectory cubic smooth interpolation.

    Args:
        traj (dict): trajectory to smooth ({motor_name: list of motor pos})
        nb_kp (int): number of keypoints to use for the cubic smoothing
        out_points (int): number of samples in the output trajectory (use -1 to conserve the same number as the input trajectory)
    """
    as_dict = isinstance(traj, dict) or isinstance(traj, np.lib.npyio.NpzFile)

    if as_dict:
        Y = np.array(list(traj.values())).T
    else:
        Y = traj

    X = np.linspace(0, 1, Y.shape[0], endpoint=True)
    fY = interp1d(X, Y.T, kind='linear')

    kx = np.linspace(0, 1, nb_kp)
    KP = fY(kx)

    if out_points == -1:
        out_points = len(X)

    C = interp1d(kx, KP, kind='cubic')
    SY = C(np.linspace(0, 1, out_points, endpoint=True)).T

    if as_dict:
        return {
            m: SY[:, i]
            for i, m in enumerate(traj.keys())
        }
    else:
        return SY


interpolation_modes = {
    'linear': Linear,
    'minjerk': MinimumJerk,
}



"""Utility toolbox submodules."""

import numpy as np

from scipy.spatial.transform import Rotation as R


def rot(axis, deg):
    """Compute 3D rotation matrix given euler rotation."""
    return R.from_euler(axis, np.deg2rad(deg)).as_matrix()



"""WebSocket IO definition."""

import json
import asyncio
import websockets
import numpy as np

from threading import Thread, Event
from base64 import b64decode
from io import BytesIO
from PIL import Image

from .io import IO


class WsIO(IO):
    """WebSocket IO implementation."""

    ws = None

    def __init__(self, part_name):
        """Init an io attached to the given part."""
        self.part_name = part_name
        self.motors = []
        self.disks = []

    @classmethod
    def shared_server(cls, part_name):
        """Create a new io with its ws server."""
        io = cls(part_name)

        if cls.ws is None:
            cls.ws = WsServer()
            cls.ws.run_in_background()
        cls.ws.register(io)

        return io

    def find_module(self, module_name):
        """Get a specific module from the IO.

        For the moment no module are really implemented. Only placeholders for code compatibility are provided.
        """
        if module_name == 'force_gripper':
            force_sensor = WsFakeForceSensor()

            if self.part_name == 'left_arm.hand':
                self.ws.left_force_sensor = force_sensor
            elif self.part_name == 'right_arm.hand':
                self.ws.right_force_sensor = force_sensor

            return force_sensor

        raise NotImplementedError

    def find_dxl(self, dxl_name, dxl_config):
        """Get a specific dynamixel motor from the IO.

        Only goal position is used atm.
        """
        pos = dxl_config['offset'] * (-1 if dxl_config['orientation'] == 'indirect' else 1)
        m = WsMotor(name=f'{self.part_name}.{dxl_name}', initial_position=pos)
        self.motors.append(m)
        self.ws.motors[m.name] = m
        return m

    def find_fan(self, fan_name):
        """Get a specific fan from its name."""
        return FakeFan()

    def find_orbita_disks(self):
        """Get a specific orbita module from the IO.

        Not currently supported.
        """
        bottomOrb = WsFakeOrbitaDisk(name=f'{self.part_name}.disk_bottom', initial_position=-60)
        middleOrb = WsFakeOrbitaDisk(name=f'{self.part_name}.disk_middle', initial_position=-60)
        topOrb = WsFakeOrbitaDisk(name=f'{self.part_name}.disk_top', initial_position=-60)

        disks = [bottomOrb, middleOrb, topOrb]
        self.disks += disks

        return disks

    def find_camera(self, index):
        """Retrieve a dual camera."""
        side = 'left' if index == 0 else 'right'
        cam = WsDualCamera(side=side)
        setattr(self.ws, f'{side}_camera', cam)
        return cam

    def close(self):
        """Close the WS."""
        self.ws.close()


class WsMotor(object):
    """Motor Placeholder.

    Only the goal position (ie. target_rot_position) is currently used.
    """

    def __init__(self, name, initial_position):
        """Init the fake motor."""
        self.name = name

        self.compliant = False
        self.target_rot_position = initial_position
        self.rot_position = initial_position
        self.temperature = 20


class WsFakeOrbitaDisk(object):
    """Orbital disk placeholder."""

    def __init__(self, name, initial_position):
        """Create fake Orbita disk."""
        self.name = name
        self.compliant = False
        self._target_rot_position = initial_position

    def __repr__(self) -> str:
        """Public Orbita disk string representation."""
        return f'<Orbita "{self.name}" pos="{self.rot_position}>'

    def setup(self):
        """Initialize the disk."""
        pass

    @property
    def rot_position(self):
        """Get the current disk angle position (in deg.)."""
        return self.target_rot_position

    @property
    def temperature(self):
        """Get the current temperature in C."""
        return 37.2

    @property
    def target_rot_position(self):
        """Get the current target angle position (in deg.)."""
        return self._target_rot_position

    @target_rot_position.setter
    def target_rot_position(self, new_pos):
        """Set a new target angle position (in deg.)."""
        self._target_rot_position = new_pos


class WsFakeForceSensor(object):
    """Force Sensor placeholder.

    Always return a nan as force.
    """

    def __init__(self):
        """Init the fake force sensor."""
        self.load = 0


class WsDualCamera(object):
    """Remote Camera."""

    def __init__(self, side):
        """Set remote camera up."""
        self.frame = np.zeros((300, 480, 3), dtype=np.uint8)

    def read(self):
        """Get latest received frame."""
        return True, self.frame

    def close(self):
        """Close the camera."""
        pass


class WsServer(object):
    """WebSocket server, sync value from the modules with their equivalent from the client."""

    def __init__(self, host='0.0.0.0', port=6171):
        """Prepare the ws server."""
        self.host, self.port = host, port
        self.running = Event()

        self.parts = []
        self.motors = {}

    async def sync(self, websocket, path):
        """Sync loop that exchange modules state with the client."""
        self.running.set()

        while self.running.is_set():
            if not websocket.open:
                break

            msg = json.dumps({
                'motors': [
                    {'name': m.name, 'goal_position': m.target_rot_position}
                    for m in sum([p.motors for p in self.parts], [])
                ],
                'disks': [
                    {'name': m.name, 'goal_position': m.target_rot_position}
                    for m in sum([p.disks for p in self.parts], [])
                ],
            })
            await websocket.send(msg.encode('UTF-8'))

            resp = await websocket.recv()
            state = json.loads(resp)

            for side in ('left', 'right'):
                eye = f'{side}_eye'

                if eye in state and hasattr(self, f'{side}_camera'):
                    jpeg_data = b64decode(state[eye])
                    frame = np.array(Image.open(BytesIO(jpeg_data)))
                    getattr(self, f'{side}_camera').frame = frame

            for m in state['motors']:
                if m['name'] in self.motors:
                    self.motors[m['name']].rot_position = m['present_position']

            if hasattr(self, 'left_force_sensor') and 'left_force_sensor' in state:
                self.left_force_sensor.load = state['left_force_sensor']
            if hasattr(self, 'right_force_sensor') and 'right_force_sensor' in state:
                self.right_force_sensor.load = state['right_force_sensor']

    def close(self):
        """Stop the sync loop."""
        self.running.clear()
        self.t.join()

    def register(self, io):
        """Register a new io (and its module) to be synced."""
        self.parts.append(io)

    def run_forever(self):
        """Run the sync loop forever."""
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)

        serve = websockets.serve(self.sync, self.host, self.port)

        loop.run_until_complete(serve)
        loop.run_forever()

    def run_in_background(self):
        """Run the sync loop forever in background."""
        self.t = Thread(target=self.run_forever)
        self.t.daemon = True
        self.t.start()


class FakeFan(object):
    """Fake fan module for API consistensy."""

    def on(self):
        """Do nothing."""
        pass

    def off(self):
        """Do nothing."""
        pass



"""Reachy top-level module.

Exposes:

* :py:mod:`~reachy.__version__` number
* :py:class:`~Reachy` class
* :py:mod:`~reachy.parts` submodule
"""

from ._version import __version__  # noqa: F401
from .reachy import Reachy  # noqa: F401



"""Trajectory player module."""

import time
import numpy as np

from threading import Thread
from operator import attrgetter


class TrajectoryPlayer(object):
    """Trajectory player abstraction.

    Args:
        reachy (:py:class:`~reachy.Reachy`): robot which will play the trajectory
        trajectories (dict): Trajectory to play (such as {motor_name: list of positions})
        freq (float): Replay sample frequency (in Hz)

    Provides high-level features to:
        * play a pre-defined trajectory
        * wait for the end of the replay
        * add fade in to smooth begining of motion
    """

    def __init__(self, reachy, trajectories, freq=100):
        """Create the Trajectory Player."""
        motor_names, trajectories = zip(*trajectories.items())

        self._reachy = reachy
        self._motors = [attrgetter(name)(reachy) for name in motor_names]
        self._traj = np.array(trajectories).T

        self._play_t = None

        self.freq = freq

    def play(self, wait=False, fade_in_duration=0):
        """Play a given trajectory.

        Args:
            wait (bool): whether or not to wait for the end of the trajectory replay
            fade_in_duration (float): time in seconds to reach the starting position (can be used to smooth the begining of the motion)

        .. warning:: you can call play multiple times to replay the trajectory. You are responisble for handling the concurrency issue if you try to run multiple trajectories on the same motors.
        """
        if fade_in_duration > 0:
            self._reachy.goto(
                goal_positions=dict(zip([m.name for m in self._motors], self._traj[0, :])),
                duration=fade_in_duration,
                starting_point='goal_position',
                wait=True,
                interpolation_mode='minjerk',
            )

        self._play_t = Thread(target=self._play_loop)
        self._play_t.start()

        if wait:
            self.wait_for_end()

    def wait_for_end(self):
        """Block until the end of a trajectory replay."""
        if self._play_t is not None and self._play_t.is_alive():
            self._play_t.join()

    def _play_loop(self):
        for pt in self._traj:
            for i, m in enumerate(self._motors):
                m.goal_position = pt[i]

            time.sleep(1 / self.freq)



"""Hand part modules.

Define different hand parts:

* a :py:class:`~reachy.parts.hand.ForceGripper`
* a :py:class:`~reachy.parts.hand.OrbitaWrist`
"""

import time
import numpy as np

from collections import OrderedDict

from ..utils import rot
from .part import ReachyPart


class Hand(ReachyPart):
    """Hand abstraction part.

    Args:
        side (str): which side the hand is attached to ('left' or 'right')
        io (str): port name where the modules can be found
    """

    fans = {}

    def __init__(self, root, io):
        """Create hand part."""
        ReachyPart.__init__(self, name=f'{root.name}.hand', io=io)

        self.side = root.side


class EmptyHand(Hand):
    """Wrist with no hand Part.

    Args:
        io (str): port name where the modules can be found
        side (str): which side the part is attached to ('left' or 'right')

    Composed of three dynamixel motors.
    """

    fans = {'wrist_fan': 'hand.wrist_pitch'}

    def __init__(self, root, io):
        """Create a new Empty Hand."""
        Hand.__init__(self, root=root, io=io)

        dxl_motors = OrderedDict({
            name: dict(conf)
            for name, conf in self.dxl_motors.items()
        })

        self.attach_dxl_motors(dxl_motors)


class LeftEmptyHand(EmptyHand):
    """Left Empty Hand Part."""

    dxl_motors = OrderedDict([
        ('forearm_yaw', {
            'id': 24, 'offset': 0.0, 'orientation': 'indirect',
            'angle-limits': [-100, 100],
            'link-translation': [0, 0, 0], 'link-rotation': [0, 0, 1],
        }),
        ('wrist_pitch', {
            'id': 25, 'offset': 0.0, 'orientation': 'indirect',
            'angle-limits': [-45, 45],
            'link-translation': [0, 0, -0.25], 'link-rotation': [0, 1, 0],
        }),
        ('wrist_roll', {
            'id': 26, 'offset': 0.0, 'orientation': 'indirect',
            'angle-limits': [-45, 45],
            'link-translation': [0, 0, -0.0325], 'link-rotation': [1, 0, 0],
        }),
    ])


class RightEmptyHand(EmptyHand):
    """Right Empty Hand Part."""

    dxl_motors = OrderedDict([
        ('forearm_yaw', {
            'id': 14, 'offset': 0.0, 'orientation': 'indirect',
            'angle-limits': [-100, 100],
            'link-translation': [0, 0, 0], 'link-rotation': [0, 0, 1],
        }),
        ('wrist_pitch', {
            'id': 15, 'offset': 0.0, 'orientation': 'indirect',
            'angle-limits': [-45, 45],
            'link-translation': [0, 0, -0.25], 'link-rotation': [0, 1, 0],
        }),
        ('wrist_roll', {
            'id': 16, 'offset': 0.0, 'orientation': 'indirect',
            'angle-limits': [-45, 45],
            'link-translation': [0, 0, -0.0325], 'link-rotation': [1, 0, 0],
        }),
    ])


class ForceGripper(Hand):
    """Force Gripper Part.

    Args:
        io (str): port name where the modules can be found
        side (str): which side the part is attached to ('left' or 'right')

    Composed of three dynamixel motors and a force sensor for gripping pressure.
    """

    fans = {'wrist_fan': 'hand.wrist_pitch'}

    def __init__(self, root, io):
        """Create a new Force Gripper Hand."""
        Hand.__init__(self, root=root, io=io)

        dxl_motors = OrderedDict({
            name: dict(conf)
            for name, conf in self.dxl_motors.items()
        })

        self.attach_dxl_motors(dxl_motors)

        self._load_sensor = self.io.find_module('force_gripper')
        self._load_sensor.offset = 4
        self._load_sensor.scale = 10000

    def __repr__(self):
        """Force gripper representation."""
        return f'<ForceGripper "grip force": {self.grip_force}>'

    def open(self, end_pos=-30, duration=1):
        """Open the gripper.

        Args:
            end_pos (float): open end position (in degrees)
            duration (float): open move duration (in seconds)
        """
        if self.side == 'left':
            end_pos = -end_pos

        self.gripper.goto(
            goal_position=end_pos,
            duration=duration,
            wait=True,
            interpolation_mode='minjerk',
        )

    def close(self, end_pos=30, duration=1, target_grip_force=100):
        """Close the gripper.

        Args:
            end_pos (float): close end position (in degrees)
            duration (float): close move duration (in seconds)
            target_grip_force (float): force threshold to stop closing the gripper

        Returns:
            bool: whether we did grasp something
        """
        release_threshold = 2

        if self.side == 'left':
            end_pos = -end_pos
            release_threshold = -release_threshold

        motion = self.gripper.goto(
            goal_position=end_pos,
            duration=duration,
            wait=False,
            interpolation_mode='minjerk',
        )

        while abs(self.grip_force) < target_grip_force and motion.is_playing:
            time.sleep(0.01)

        motion.stop()
        time.sleep(0.1)

        self.gripper.goal_position = self.gripper.present_position - release_threshold
        time.sleep(0.5)

        return abs(self.grip_force) > 0.5 * target_grip_force

    @property
    def grip_force(self):
        """Get current grip force."""
        return self._load_sensor.load


class LeftForceGripper(ForceGripper):
    """Left Force Gripper Part."""

    dxl_motors = OrderedDict([
        ('forearm_yaw', {
            'id': 24, 'offset': 0.0, 'orientation': 'indirect',
            'angle-limits': [-100, 100],
            'link-translation': [0, 0, 0], 'link-rotation': [0, 0, 1],
        }),
        ('wrist_pitch', {
            'id': 25, 'offset': 0.0, 'orientation': 'indirect',
            'angle-limits': [-45, 45],
            'link-translation': [0, 0, -0.25], 'link-rotation': [0, 1, 0],
        }),
        ('wrist_roll', {
            'id': 26, 'offset': 0.0, 'orientation': 'indirect',
            'angle-limits': [-45, 45],
            'link-translation': [0, 0, -0.0325], 'link-rotation': [1, 0, 0],
        }),
        ('gripper', {
            'id': 27, 'offset': 0.0, 'orientation': 'direct',
            'angle-limits': [-20, 69],
            'link-translation': [0, -0.01, -0.075], 'link-rotation': [0, 0, 0],
        }),
    ])


class RightForceGripper(ForceGripper):
    """Right Force Gripper Part."""

    dxl_motors = OrderedDict([
        ('forearm_yaw', {
            'id': 14, 'offset': 0.0, 'orientation': 'indirect',
            'angle-limits': [-100, 100],
            'link-translation': [0, 0, 0], 'link-rotation': [0, 0, 1],
        }),
        ('wrist_pitch', {
            'id': 15, 'offset': 0.0, 'orientation': 'indirect',
            'angle-limits': [-45, 45],
            'link-translation': [0, 0, -0.25], 'link-rotation': [0, 1, 0],
        }),
        ('wrist_roll', {
            'id': 16, 'offset': 0.0, 'orientation': 'indirect',
            'angle-limits': [-45, 45],
            'link-translation': [0, 0, -0.0325], 'link-rotation': [1, 0, 0],
        }),
        ('gripper', {
            'id': 17, 'offset': 0.0, 'orientation': 'direct',
            'angle-limits': [-69, 20],
            'link-translation': [0, -0.01, -0.075], 'link-rotation': [0, 0, 0],
        }),
    ])


class OrbitaWrist(Hand):
    """Orbita Wrist hand.

    Args:
        io (str): port name where the modules can be found
        side (str): which side the part is attached to ('left' or 'right')

    A 3dof Orbita Wrist at the end of the arm.
    """

    dxl_motors = {}
    orbita_config = {
        'Pc_z': [0, 0, 25],
        'Cp_z': [0, 0, 0],
        'R': 36.7,
        'R0': rot('z', 60),
        'hardware_zero': np.zeros(3),
    }

    def __init__(self, root, io):
        """Create a new OrbitaWrist Hand."""
        Hand.__init__(self, root=root, io=io)

        self.wrist = self.create_orbita_actuator('wrist', OrbitaWrist.orbita_config)

    def __repr__(self):
        """Orbita wrist representation."""
        return f'<OrbitaWrist "wrist": {self.wrist}>'

    def homing(self):
        """Launch Wrist homing procedure."""
        self.wrist.homing()



"""Configure gate and three motor modules for an Orbita Actuator.

This tool:
* change the name of the gate (eg. 'r_head')
* change the name of the controlled motor modules ('disk_top', 'disk_middle', 'disk_bottom')

"""

import time
import argparse

from contextlib import closing
from pyluos import Device as LuosDevice


def main():
    """Run main orbita config entry point."""
    parser = argparse.ArgumentParser()
    parser.add_argument('luos_port')
    parser.add_argument('part_name', choices=['head'])
    args = parser.parse_args()

    with closing(LuosDevice(args.luos_port)) as io:
        if len(io.modules) != 4:
            names = [m.alias for m in io.modules]
            raise ValueError(f'Incoherent modules found: {names}!')

        gate = io.modules[0]
        gate_name = f'r_{args.part_name}'
        if gate.alias != gate_name:
            print(f'Changing gate "{gate.alias}" name to "{gate_name}"')
            gate.rename(gate_name)
            time.sleep(0.5)

        for module, disk in zip(io.modules[1:], ['top', 'middle', 'bottom']):
            disk_name = f'disk_{disk}'
            if module.alias != disk_name:
                print(f'Changing module "{module.alias}" to "{disk_name}"')
                module.rename(disk_name)
                time.sleep(0.5)

    print('Everything is setup!')


if __name__ == '__main__':
    main()



"""Retrieve Orbita Zero hardware and write it to the specified config file.

This tool must be run while Orbita is fixed at the Zero Position using a specific 3D part.

"""

import time
import pathlib
import argparse
import numpy as np

from glob import glob

from pyluos import Device


import reachy


def main():
    """Get the Zero hardware from Orbita."""
    output_filename = pathlib.Path(reachy.__file__).parent / 'orbita_head_hardware_zero.npy'

    parser = argparse.ArgumentParser()
    parser.add_argument('--luos_port', default='/dev/ttyUSB*', help='Orbita gate luos port (default: %(default)s)')
    parser.add_argument('--output_filename', default=output_filename, help='output file where to store the zero (default: %(default)s)')
    args = parser.parse_args()

    ports = glob(args.luos_port)
    if len(ports) == 0:
        ports = [args.luos_port]

    device = Device(ports[0])
    time.sleep(0.5)

    hardware_zero = (
        device.disk_top.rot_position,
        device.disk_middle.rot_position,
        device.disk_bottom.rot_position,
    )

    print(f'Find Orbita Hardware Zero at {hardware_zero}')
    np.save(args.output_filename, hardware_zero)


if __name__ == '__main__':
    main()



"""Trajectory recording utility module."""

import time
import numpy as np

from threading import Event, Thread


class TrajectoryRecorder(object):
    """Trajectory Recorder utility class.

    Args:
        motors (list): list of motors to record (eg. [reachy.right_arm.elbow_pitch, reachy.right_arm.shoulder_pitch])
        position_field (str): register to record as trajectories (default use the 'present_position', 'goal_position' can also be useful in some specific case)
        freq (float): record sample frequency (in Hz)

    .. note:: A same recorder can be used to record multiple trajectories.

    Facilitates the recording of a full trajectory on multiple motors.
    """

    def __init__(self, motors, position_field='present_position', freq=100):
        """Create the recorder."""
        self.motors = motors

        self._data = []
        self._position_field = position_field

        self._recording = Event()
        self._record_t = None

        self.freq = freq

    def start(self, turn_compliant=False):
        """Start the record.

        Args:
            turn_compliant (bool): whether or not to turn the motor compliant before starting the record.
        """
        self._recording.set()
        self._record_t = Thread(target=self._record_loop)

        if turn_compliant:
            for m in self.motors:
                m.compliant = True

        self._record_t.start()

    def stop(self, turn_stiff=False):
        """Stop the record.

        Args:
            turn_stiff (bool): whether or not to turn the motor stiff at the end of the record.
        """
        self._recording.clear()
        if self._record_t is not None and self._record_t.is_alive():
            self._record_t.join()

        if turn_stiff:
            for m in self.motors:
                m.compliant = False

    @property
    def trajectories(self):
        """Retrieve a copy of the recorded trajectories."""
        traj = np.array(self._data).T

        return {
             self.motors[i].name: t
             for i, t in enumerate(traj)
         }

    def _record_loop(self):
        self._data.clear()

        while self._recording.is_set():
            self._data.append([
                getattr(m, self._position_field)
                for m in self.motors
            ])

            time.sleep(1 / self.freq)



"""Command line utility tool for configuring dynamixel motor.

This tool let you configure one of the dynamixel motor used in Reachy. It can be used when you have to change a broken motor for instance.
In more details, this tool will:

* change the id
* set the angle limits

.. warning:: Make sure to have only one dynamixel motor connected when you configure it!
"""

import time
import argparse

from contextlib import closing
from pyluos import Device as LuosDevice

from reachy.parts import RightArm, LeftArm, Head
from reachy.parts.hand import LeftForceGripper, RightForceGripper

reachy_config = {}
reachy_config.update({f'right_arm.{m}': c for m, c in RightArm.dxl_motors.items()})
reachy_config.update({f'right_arm.hand.{m}': c for m, c in RightForceGripper.dxl_motors.items()})
reachy_config.update({f'left_arm.{m}': c for m, c in LeftArm.dxl_motors.items()})
reachy_config.update({f'left_arm.hand.{m}': c for m, c in LeftForceGripper.dxl_motors.items()})
reachy_config.update({f'head.{m}': c for m, c in Head.dxl_motors.items()})


def main():
    """Run main dxl-config entry point."""
    parser = argparse.ArgumentParser()
    parser.add_argument('luos_port')
    parser.add_argument('motor_name', choices=reachy_config.keys())
    args = parser.parse_args()

    name = args.motor_name
    conf = reachy_config[name]
    id = conf['id']
    limits = conf['angle-limits']

    with closing(LuosDevice(args.luos_port)) as io:
        print(f'Will configure motor {name}...')

        dxl_motors_found = [m for m in io.modules if m.alias.startswith('dxl_')]

        if len(dxl_motors_found) == 0:
            print('No motor found with 1M baudrate. Trying to setup the correct baudrate...')
            io.void_dxl.baudrate = 1000000
            time.sleep(0.5)
            print('Unplug/Replug the modules and power supply and try again.')

        if len(dxl_motors_found) > 1:
            raise EnvironmentError(f'{len(dxl_motors_found)} motors found on {args.luos_port}! Connect only the motor you want to configure and try again!')

        dxl_motor = dxl_motors_found[0]

        print(f'\t with id {id}...')
        dxl_motor.set_id(conf['id'])
        time.sleep(0.25)

        print(f'\t and limits {limits}...')
        dxl_motor.rot_position_limit = conf['angle-limits']

        time.sleep(0.25)
        print('Done!')

    print('Ok!')


if __name__ == '__main__':
    main()
